/*
 * linux/arch/m68k/console/cvision.c -- Low level text mode based
 *                                      CyberVision64 console.
 *
 *    Copyright (C) 1997 Antonio Santos
 *
 *
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file COPYING in the main directory of this archive
 * for more details.
 */


#include <linux/types.h>
#include <linux/console.h>


#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/tty.h>
#include <linux/malloc.h>
#include <linux/delay.h>
#include <asm/uaccess.h>
#include <asm/system.h>
#include <asm/irq.h>
#include <linux/zorro.h>
#include <asm/pgtable.h>
#include <linux/fb.h>

#define CVISION_C
#include "cvision.h"
#undef CVISION_C

/* imported from cvision_core.c */
extern long cv64_memclk;


/* ----------------------------------------------------------- */

#define arraysize(x)			(sizeof(x)/sizeof(*(x)))

struct cvision_fb_par {
	int x_resolution;
	int y_resolution;
	int bits_per_pixel;
};


/* ----------------------------------------------------------- */

	/* Interface used by the world */
int		cvision_probe 			(void);
void	cvision_video_setup 	(char *options, int *ints);

int		cvision_fb_get_fix		(struct fb_fix_screeninfo *fix, int con);
int		cvision_fb_get_var		(struct fb_var_screeninfo *var, int con);
int		cvision_fb_set_var		(struct fb_var_screeninfo *var, int con);
int		cvision_fb_get_cmap		(struct fb_cmap *cmap, int kspc, int con);
int		cvision_fb_set_cmap		(struct fb_cmap *cmap, int kspc, int con);
int		cvision_fb_pan_display	(struct fb_var_screeninfo *var, int con);
int		cvision_fb_ioctl		(struct inode *inode, struct file *file, u_int cmd, u_long arg, int con);

	/* Interface to the low level console driver */
struct fb_info *	cvisionfb_init		(long *mem_start);	/* Through amiga_fb_init () */
int					cvisionfb_switch	(int con);
int					cvisionfb_updatevar	(int con);
void				cvisionfb_blank		(int blank);
int					cvisionfb_setcmap	(struct fb_cmap *cmap, int con);

	/* Accelerated functions used by the low level console driver */
				/* --- these are in cvision_core.c --- */
		
	/* Hardware specific routines */
int		cvision_init (void);
int		cvision_encode_fix	(struct fb_fix_screeninfo *fix, struct cvision_fb_par *par);
int		cvision_decode_var	(struct fb_var_screeninfo *var, struct cvision_fb_par *par);
int		cvision_encode_var	(struct fb_var_screeninfo *var,	struct cvision_fb_par *par);
int		cvision_getcolreg	(u_int regno, u_int *red, u_int *green, u_int *blue, u_int *transp);
int		cvision_setcolreg	(u_int regno, u_int red, u_int green, u_int blue, u_int transp);
void	cvision_blank		(int blank);

	/* Internal routines */
void					cvision_fb_get_par		(struct cvision_fb_par *par);
void					cvision_fb_set_par		(struct cvision_fb_par *par);
static int				do_fb_set_var			(struct fb_var_screeninfo *var, int isactive);
void					cvision_fb_set_disp		(int con);
static int				get_video_mode			(const char *name);
static struct fb_cmap *	get_default_colormap	(int bits_per_pixel);
static int				do_fb_get_cmap			(struct fb_cmap *cmap, struct fb_var_screeninfo *var, int kspc);
static int				do_fb_set_cmap			(struct fb_cmap *cmap, struct fb_var_screeninfo *var, int kspc);
static void				do_install_cmap			(int con);
static void				memcpy_fs				(int fsfromto, void *to, void *from, int len);
static void				copy_cmap				(struct fb_cmap *from, struct fb_cmap *to, int fsfromto);
static int				alloc_cmap				(struct fb_cmap *cmap, int len, int transp);




struct cvision_fb_par	current_par;
int						current_par_valid = 0;
int						currcon = 0;
struct display			disp [MAX_NR_CONSOLES];
struct fb_info			fb_info;
int						node;						/* node of the /dev/fb?current file */
char					cvision_fb_name[16] = "CyberVision64";

/* ---------------------------------------------------------------------------- */

struct fb_hwswitch {
	/* - initialisation - */
	int (*init) (void);
	
	/* - display control - */
	int (*encode_fix) (struct fb_fix_screeninfo *fix, struct cvision_fb_par *par);
	int (*decode_var) (struct fb_var_screeninfo *var, struct cvision_fb_par *par);
	int (*encode_var) (struct fb_var_screeninfo *var, struct cvision_fb_par *par);
	int (*getcolreg) (u_int regno, u_int *red, u_int *green, u_int *blue, u_int *transp);
	int (*setcolreg) (u_int regno, u_int red, u_int green, u_int blue, u_int transp);
	void (*blank) (int blank);
} *fbhw;

int cvision_key = 0;
u_char cvision_colour_table [256][4];
unsigned long cv64_mem;
unsigned long cv64_regs;
unsigned long cv64_fbmem;
unsigned long cv64_size;
long *memstart;

char *cvision_fb_modenames [] = {
	/* Default Video Mode (640x480x8) */
	"default",
	
	/* Predefined Video Modes */
	"cvision640x480x8",
	"cvision800x600x8",
	"cvision1024x768x8",
	"cvision1152x886x8",
	"cvision1280x1024x8i",
	"dummy", "dummy", "dummy", "dummy", "dummy", "dummy", "dummy", "dummy",
	
	/* User Defined Video Modes */
	"user0", "user1", "user2", "user3", "user4", "user5", "user6", "user7"
};

struct fb_var_screeninfo cvision_fb_predefined [] = {
		/* Default Video Mode (640x480x8) */
	{
		640, 480, 640, 480, 0, 0, 8, 0,					/* Width, Height, Virtual Width, Virtual Height, Xoffset, Yoffset, Bits Per Pixel, GrayScale ? */
		{0, 8, 0}, {0, 8, 0}, {0, 8, 0}, {0, 0, 0},		/* bitfields: red, green, blue, transp */
		0, 0,											/* nonstd, activate */
		-1, -1,											/* Height in mm, Width in mm */
		FB_ACCEL_CVISION,								/* accel */
		39722,											/* pixel clock in pico seconds */
		40, 24, 32, 11, 96, 2,							/* left_margin, right_margin, upper_margin, lower_margin, hs, vs */
		FB_SYNC_COMP_HIGH_ACT | FB_SYNC_VERT_HIGH_ACT,	/* sync */
		FB_VMODE_NONINTERLACED							/* vmode */
	},
		/* 640x480x8 Video Mode */
	{
		640, 480, 640, 480, 0, 0, 8, 0,					/* Width, Height, Virtual Width, Virtual Height, Xoffset, Yoffset, Bits Per Pixel, GrayScale ? */
		{0, 8, 0}, {0, 8, 0}, {0, 8, 0}, {0, 0, 0},		/* bitfields: red, green, blue, transp */
		0, 0,											/* nonstd, activate */
		-1, -1,											/* Height in mm, Width in mm */
		FB_ACCEL_CVISION,								/* accel */
		39722,											/* pixel clock in pico seconds */
		40, 24, 32, 11, 96, 2,							/* left_margin, right_margin, upper_margin, lower_margin, hs, vs */
		FB_SYNC_COMP_HIGH_ACT | FB_SYNC_VERT_HIGH_ACT,	/* sync */
		FB_VMODE_NONINTERLACED							/* vmode */
	},
		/* 800x600x8 Video Mode */
	{
		800, 600, 800, 600, 0, 0, 8, 0,					/* Width, Height, Virtual Width, Virtual Height, Xoffset, Yoffset, Bits Per Pixel, GrayScale ? */
		{0, 8, 0}, {0, 8, 0}, {0, 8, 0}, {0, 0, 0},		/* bitfields: red, green, blue, transp */
		0, 0,											/* nonstd, activate */
		-1, -1,											/* Height in mm, Width in mm */
		FB_ACCEL_CVISION,								/* accel */
		27778,											/* pixel clock in pico seconds */
		64, 24, 22, 1, 72, 2,							/* left_margin, right_margin, upper_margin, lower_margin, hs, vs */
		FB_SYNC_COMP_HIGH_ACT | FB_SYNC_VERT_HIGH_ACT,	/* sync */
		FB_VMODE_NONINTERLACED							/* vmode */
	},

		/* 1024x768x8 Video Mode */
	{
		1024, 768, 1024, 768, 0, 0, 8, 0,				/* Width, Height, Virtual Width, Virtual Height, Xoffset, Yoffset, Bits Per Pixel, GrayScale ? */
		{0, 8, 0}, {0, 8, 0}, {0, 8, 0}, {0, 0, 0},		/* bitfields: red, green, blue, transp */
		0, 0,											/* nonstd, activate */
		-1, -1,											/* Height in mm, Width in mm */
		FB_ACCEL_CVISION,								/* accel */
		16667,											/* pixel clock in pico seconds */
		224, 72, 60, 12, 168, 4,						/* left_margin, right_margin, upper_margin, lower_margin, hs, vs */
		FB_SYNC_COMP_HIGH_ACT | FB_SYNC_VERT_HIGH_ACT,	/* sync */
		FB_VMODE_NONINTERLACED							/* vmode */
	},
		/* 1152x886x8 Video Mode */
	{
		1152, 886, 1152, 886, 0, 0, 8, 0,				/* Width, Height, Virtual Width, Virtual Height, Xoffset, Yoffset, Bits Per Pixel, GrayScale ? */
		{0, 8, 0}, {0, 8, 0}, {0, 8, 0}, {0, 0, 0},		/* bitfields: red, green, blue, transp */
		0, 0,											/* nonstd, activate */
		-1, -1,											/* Height in mm, Width in mm */
		FB_ACCEL_CVISION,								/* accel */
		15873,											/* pixel clock in pico seconds */
		184, 40, 24, 1, 56, 16,							/* left_margin, right_margin, upper_margin, lower_margin, hs, vs */
		FB_SYNC_COMP_HIGH_ACT | FB_SYNC_VERT_HIGH_ACT,	/* sync */
		FB_VMODE_NONINTERLACED							/* vmode */
	},
		/* 1280x1024x8i Video Mode */
	{
		1280, 1024, 1280, 1024, 0, 0, 8, 0,				/* Width, Height, Virtual Width, Virtual Height, Xoffset, Yoffset, Bits Per Pixel, GrayScale ? */
		{0, 8, 0}, {0, 8, 0}, {0, 8, 0}, {0, 0, 0},		/* bitfields: red, green, blue, transp */
		0, 0,											/* nonstd, activate */
		-1, -1,											/* Height in mm, Width in mm */
		FB_ACCEL_CVISION,								/* accel */
		16667,											/* pixel clock in pico seconds */
		256, 48, 50, 12, 72, 4,							/* left_margin, right_margin, upper_margin, lower_margin, hs, vs */
		FB_SYNC_COMP_HIGH_ACT | FB_SYNC_VERT_HIGH_ACT,	/* sync */
		FB_VMODE_INTERLACED								/* vmode */
	},
		/* Dummy Video Modes */
	{0, }, {0, }, {0, }, {0, }, {0, }, {0, }, {0, }, {0, },
		/* User Defined Video Modes */
	{0, }, {0, }, {0, }, {0, }, {0, }, {0, }, {0, }, {0, }
};

#define NUM_TOTAL_MODES		(arraysize(cvision_fb_predefined))
#define NUM_PREDEF_MODES	(6)

int cvision_fb_inverse = 0;
int cvision_fb_use = 0;			/* User CVision driver ? */
int cvision_fb_mode = 0;


int cvision_custom_mode = 0,
	hbs, hbe, hss, hse, ht,
	vbs, vbe, vss, vse, vt;



struct fb_hwswitch cvision_switch = {
	cvision_init, cvision_encode_fix, cvision_decode_var, cvision_encode_var,
	cvision_getcolreg, cvision_setcolreg, cvision_blank
};

/* default colormaps */

u_short red16 [] = {
	0x0000, 0x0000, 0x0000, 0x0000, 0xC000, 0xC000, 0xC000, 0xC000,
	0x8000, 0x0000, 0x0000, 0x0000, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF
};

u_short green16 [] = {
	0x0000, 0x0000, 0xC000, 0xC000, 0x0000, 0x0000, 0xC000, 0xC000,
	0x8000, 0x0000, 0xFFFF, 0xFFFF, 0x0000, 0x0000, 0xFFFF, 0xFFFF
};

u_short blue16 [] = {
	0x0000, 0x0000, 0x0000, 0xC000, 0x0000, 0xC000, 0x0000, 0xC000,
	0x8000, 0xFFFF, 0x0000, 0xFFFF, 0x0000, 0xFFFF, 0x0000, 0xFFFF
};

struct fb_cmap default_16_colors = {
	0, 16, red16, green16, blue16, NULL
};

#define CNVT_TOHW(val, width)	((((val)<<(width))+0x7FFF-(val))>>16)
#define CNVT_FROMHW(val, width)	(((width)?((((val)<<16)-(val))/((1<<(width))-1)):0))

struct fb_ops cvision_fb_ops = {
	cvision_fb_get_fix, cvision_fb_get_var, cvision_fb_set_var, cvision_fb_get_cmap,
	cvision_fb_set_cmap, cvision_fb_pan_display, cvision_fb_ioctl
};



/* ----------------------------------------------------------- */
int cvision_probe (void)
{
	cvision_key = zorro_find (MANUF_PHASE5, PROD_CYBERVISION, 0, 0);
	return (cvision_key);	
}

void cvision_video_setup (char *options, int *ints)
{
	char *o;
	int i;
	char custom_mode[256];
	
	
	custom_mode [0] = '\0';
	fb_info.fontname [0] = '\0';
	
	if (!options || !*options)
		return;
	
	for (o = strtok (options, ","); o != NULL; o = strtok (NULL, ",")) {
		if (!strcmp (o, "inverse")) {
			
			cvision_fb_inverse = 1;
			
			for (i = 0; i < 16; i++) {
				  red16 [i] =   ~red16 [i];
				green16 [i] = ~green16 [i];
				 blue16 [i] =  ~blue16 [i];
			}

		} else if (!strncmp (o, "font:", 5)) {
			strcpy (fb_info.fontname, o + 5);
		} else if (!strncmp (o, "cvision", 7)) {
			cvision_fb_use = 1;
		} else if (!strncmp (o, "custom:", 7)) {
			strcpy (custom_mode, o+7);
		}
		
		if (*custom_mode) {
			/* custom:<xres>;<yres>;<clock>;<hbs>;<hbe>;<hss>;<hse>;<ht>;<vbs>;<vbe>;<vss>;<vse>;<vt> */
		
			struct fb_var_screeninfo *si;
			char *p;

			si = &cvision_fb_predefined [cvision_fb_mode];
			
			cvision_custom_mode = 1;
			
			if (!(p = strtok (custom_mode, ";")) || !*p) goto custom_invalid;
			si->xres = si->xres_virtual = simple_strtoul (p, NULL, 10);
			if (!(p = strtok (NULL, ";")) || !*p) goto custom_invalid;
			si->yres = si->yres_virtual = simple_strtoul (p, NULL, 10);
			if (!(p = strtok (NULL, ";")) || !*p) goto custom_invalid;
			si->pixclock = simple_strtoul (p, NULL, 10);
			if (!(p = strtok (NULL, ";")) || !*p) goto custom_invalid;
			hbs = (int) simple_strtoul (p, NULL, 10);
			if (!(p = strtok (NULL, ";")) || !*p) goto custom_invalid;
			hbe = (int) simple_strtoul (p, NULL, 10);
			if (!(p = strtok (NULL, ";")) || !*p) goto custom_invalid;
			hss = (int) simple_strtoul (p, NULL, 10);
			if (!(p = strtok (NULL, ";")) || !*p) goto custom_invalid;
			hse = (int) simple_strtoul (p, NULL, 10);
			if (!(p = strtok (NULL, ";")) || !*p) goto custom_invalid;
			ht = (int) simple_strtoul (p, NULL, 10);
			if (!(p = strtok (NULL, ";")) || !*p) goto custom_invalid;
			vbs = (int) simple_strtoul (p, NULL, 10);
			if (!(p = strtok (NULL, ";")) || !*p) goto custom_invalid;
			vbe = (int) simple_strtoul (p, NULL, 10);
			if (!(p = strtok (NULL, ";")) || !*p) goto custom_invalid;
			vss = (int) simple_strtoul (p, NULL, 10);
			if (!(p = strtok (NULL, ";")) || !*p) goto custom_invalid;
			vse = (int) simple_strtoul (p, NULL, 10);
			if (!(p = strtok (NULL, "")) || !*p) goto custom_invalid;
			vt = (int) simple_strtoul (p, NULL, 10);
			
			custom_invalid:
				;
		}
		
		if (!strncmp (o, "cvision", 7)) {
			cvision_fb_mode = get_video_mode (o);
		}
	}
}

int cvision_fb_get_fix (struct fb_fix_screeninfo *fix, int con)
{
	int error = 0;
	struct cvision_fb_par par;
	
	if (con == -1) {
		cvision_fb_get_par (&par);
	} else {
		error = fbhw->decode_var (&disp [con].var, &par);
	}
	
	return (error ? error : fbhw->encode_fix (fix, &par));
}

int cvision_fb_get_var (struct fb_var_screeninfo *var, int con)
{
	int error = 0;
	struct cvision_fb_par par;
	
	if (con == -1) {
		cvision_fb_get_par (&par);
		error = fbhw->encode_var (var, &par);
	} else {
		*var = disp [con].var;
	}
	
	return (error);
}

int cvision_fb_set_var (struct fb_var_screeninfo *var, int con)
{
	int err, oldxres, oldyres, oldvxres, oldvyres, oldbpp;
	
	if ((err = do_fb_set_var (var, con == currcon))) {
		return (err);
	}
	
	if ((var->activate & FB_ACTIVATE_MASK) == FB_ACTIVATE_NOW) {
		 oldxres = disp [con].var.xres;
		 oldyres = disp [con].var.yres;
		oldvxres = disp [con].var.xres_virtual;
		oldvyres = disp [con].var.yres_virtual;
		  oldbpp = disp [con].var.bits_per_pixel;
		disp [con].var = *var;
		
		if (oldxres != var->xres || oldyres != var->yres || oldvxres != var->xres_virtual || oldvyres != var->yres_virtual || oldbpp != var->bits_per_pixel) {
			cvision_fb_set_disp (con);
			(*fb_info.changevar) (con);
			alloc_cmap (&disp [con].cmap, 0, 0);
			do_install_cmap (con);
		}
	}
	
	var->activate = 0;
	
	return (0);
}

int cvision_fb_get_cmap (struct fb_cmap *cmap, int kspc, int con)
{
	if (con == currcon) {
		return (do_fb_get_cmap (cmap, &disp [con].var, kspc));
	} else if (disp [con].cmap.len) {
		copy_cmap (&disp [con].cmap, cmap, kspc ? 0 : 2);
	} else {
		copy_cmap (get_default_colormap (disp [con].var.bits_per_pixel), cmap, kspc ? 0 : 2);
	}
	
	return (0);
}

int cvision_fb_set_cmap (struct fb_cmap *cmap, int kspc, int con)
{
	int err;
	
	if (!disp [con].cmap.len) {
		if ((err = alloc_cmap (&disp [con].cmap, 1 << disp [con].var.bits_per_pixel, 0))) {
			return (err);
		}
	}
	
	if (con == currcon) {
		return (do_fb_set_cmap (cmap, &disp [con].var, kspc));
	} else {
		copy_cmap (cmap, &disp [con].cmap, kspc ? 0 : 1);
	}
	
	return (0);
}

int cvision_fb_pan_display (struct fb_var_screeninfo *var, int con)
{
	return (-EINVAL);
}

int cvision_fb_ioctl (struct inode *inode, struct file *file, u_int cmd, u_long arg, int con)
{
	return (-EINVAL);
}

/* ----------------------------------------------------------- */

struct fb_info *cvisionfb_init (long *mem_start)
{
	int err;
	struct cvision_fb_par par;
	
	memstart = mem_start;
	
	fbhw = &cvision_switch;
	
	err = register_framebuffer (cvision_fb_name, &node, &cvision_fb_ops, NUM_TOTAL_MODES, cvision_fb_predefined);
	
	if (err < 0) {
		panic ("Cannot register frame buffer\n");
	}
	
	fbhw->init ();
	fbhw->decode_var (&cvision_fb_predefined [cvision_fb_mode], &par);
	fbhw->encode_var (&cvision_fb_predefined [0], &par);
	
	strcpy (fb_info.modename, cvision_fb_name);
	fb_info.disp = disp;
	fb_info.switch_con = &cvisionfb_switch;
	fb_info.updatevar = &cvisionfb_updatevar;
	fb_info.blank = &cvisionfb_blank;
	fb_info.setcmap = &cvisionfb_setcmap;
	
	do_fb_set_var (&cvision_fb_predefined [0], 1);
	cvision_fb_get_var (&disp [0].var, -1);
	cvision_fb_set_disp (-1);
	do_install_cmap (0);
	return (&fb_info);
}

int cvisionfb_switch (int con)
{
	if (disp [currcon].cmap.len) {
		do_fb_get_cmap (&disp [currcon].cmap, &disp [currcon].var, 1);
	}
	
	do_fb_set_var (&disp [con].var, 1);
	currcon = con;
	do_install_cmap (con);
	return (0);
}

int cvisionfb_updatevar (int con)
{
	return (0);
}

void cvisionfb_blank (int blank)
{
	fbhw->blank (blank);
}

int cvisionfb_setcmap (struct fb_cmap *cmap, int con)
{
	return (cvision_fb_set_cmap (cmap, 1, con));
}

/* ----------------------------------------------------------- */

/*
 *	Initialization.
 *
 *	Set the default video mode for this chipset. If a video mode was
 *	sepcified on the command line, it will override the default mode.
 *
 */
 
int cvision_init (void)
{
	unsigned long board_addr, board_size;
	struct ConfigDev *cd;
	
	cd = (struct ConfigDev *) zorro_get_board (cvision_key);
	zorro_config_board (cvision_key, 0);
	board_addr = (unsigned long) cd->cd_BoardAddr;
	board_size = (unsigned long) cd->cd_BoardSize;
	
	*memstart = (*memstart + PAGE_SIZE - 1) & ~(PAGE_SIZE - 1);

	cv64_mem = kernel_map (board_addr, board_size, KERNELMAP_NOCACHE_SER, memstart);
	cv64_regs = cv64_mem + 0x02000000;
	cv64_fbmem = cv64_mem + 0x01400000;
	
	cv64_board_init ();	/* here the size of board is get
	                     * and video is enabled
	                     *
	                     */
	
	/* clean up frame buffer memory */
	memset ((char *) cv64_fbmem, 0, cv64_size);

	/* load default video mode */
	cv64_load_video_mode (&cvision_fb_predefined [cvision_fb_mode]);

	return (0);
}

int cvision_encode_fix (struct fb_fix_screeninfo *fix, struct cvision_fb_par *par)
{
	int i;
	
	strcpy (fix->id, cvision_fb_name);

	fix->smem_start = cv64_fbmem;
	fix->smem_len = cv64_size;

	fix->type = FB_TYPE_PACKED_PIXELS;
	fix->type_aux = 0;

	if (par->bits_per_pixel == 15 || par->bits_per_pixel == 16 || par->bits_per_pixel == 24 || par->bits_per_pixel == 32) {
		fix->visual = FB_VISUAL_DIRECTCOLOR;
	} else {
		fix->visual = FB_VISUAL_PSEUDOCOLOR;
	}

	fix->xpanstep = 0;
	fix->ypanstep = 0;
	fix->ywrapstep = 0;
	fix->line_length = 0;
	
	for (i = 0; i < arraysize (fix->reserved); i++) {
		fix->reserved [i] = 0;
	}

	return (0);
}

int cvision_decode_var (struct fb_var_screeninfo *var, struct cvision_fb_par *par)
{
	  par->x_resolution = var->xres;
	  par->y_resolution = var->yres;
	par->bits_per_pixel = var->bits_per_pixel;
	
	return (0);
}

int cvision_encode_var (struct fb_var_screeninfo *var,	struct cvision_fb_par *par)
{
	int i;
	
	var->xres = par->x_resolution;
	var->yres = par->y_resolution;
	var->xres_virtual = par->x_resolution;
	var->yres_virtual = par->y_resolution;
	var->xoffset = 0;
	var->yoffset = 0;
	var->bits_per_pixel = par->bits_per_pixel;
	var->grayscale = 0;
	
	switch (par->bits_per_pixel) {
		case 8:
			var->red.offset = 0;
			var->red.length = 8;
			var->red.msb_right = 0;
			var->blue = var->green = var->red;
			var->pixclock = 39722;
			break;
		
		case 15:
			var->red.offset = 10;
			var->red.length = 5;
			var->red.msb_right = 0;
			var->green.offset = 5;
			var->green.length = 5;
			var->green.msb_right = 0;
			var->blue.offset = 0;
			var->blue.length = 5;
			var->blue.msb_right = 0;
			var->pixclock = 16667;
			break;
		
		case 16:
			var->red.offset = 11;
			var->red.length = 5;
			var->red.msb_right = 0;
			var->green.offset = 5;
			var->green.length = 6;
			var->green.msb_right = 0;
			var->blue.offset = 0;
			var->blue.length = 5;
			var->blue.msb_right = 0;
			var->pixclock = 16667;
			break;
		
		case 24:
		case 32:
			var->red.offset = 16;
			var->red.length = 8;
			var->red.msb_right = 0;
			var->green.offset = 8;
			var->green.length = 8;
			var->green.msb_right = 0;
			var->blue.offset = 0;
			var->blue.length = 8;
			var->blue.msb_right = 0;
			var->pixclock = 16667;
			break;
	}
	
	var->transp.offset = 0;
	var->transp.length = 0;
	var->transp.msb_right = 0;
	
	var->nonstd = 0;
	var->activate = 0;
	
	var->height = -1;
	var->width = -1;
	var->accel = FB_ACCEL_CVISION;
	var->vmode = FB_VMODE_NONINTERLACED;
	
	/* dummy values */
	
	/* var->pixclock */
	
	var->sync = 0;
	var->left_margin = 40;
	var->right_margin = 24;
	var->upper_margin = 32;
	var->lower_margin = 11;
	var->hsync_len = 96;
	var->vsync_len = 2;
	
	for (i = 0; i < arraysize (var->reserved); i++) {
		var->reserved [i] = 0;
	}
	
	return (0);
}

int cvision_getcolreg (u_int regno, u_int *red, u_int *green, u_int *blue, u_int *transp)
{
	if (regno > 255) {
		return (1);
	}
	
	   *red = cvision_colour_table [regno][0];
	 *green = cvision_colour_table [regno][1];
	  *blue = cvision_colour_table [regno][2];
	*transp = cvision_colour_table [regno][3];
	
	return (0);
}

int cvision_setcolreg (u_int regno, u_int red, u_int green, u_int blue, u_int transp)
{
	if (regno > 255) {
		return (1);
	}
	
	cvision_colour_table [regno][0] = red & 0xFF;
	cvision_colour_table [regno][1] = green & 0xFF;
	cvision_colour_table [regno][2] = blue & 0xFF;
	cvision_colour_table [regno][3] = transp;
	
	vgaw (cv64_regs, VDAC_ADDRESS_W, regno);
	vgaw (cv64_regs, VDAC_DATA, (red & 0xFF) >> 2);
	vgaw (cv64_regs, VDAC_DATA, (green & 0xFF) >> 2);
	vgaw (cv64_regs, VDAC_DATA, (blue & 0xFF) >> 2);
	
	return (0);
}

void cvision_blank (int blank)
{
	gfx_on_off (blank ? 1 : 0, (volatile unsigned char *) cv64_regs);
}


/* ----------------------------------------------------------- */

/* ----------------------------------------------------------- */
/* internal routines */

void cvision_fb_get_par (struct cvision_fb_par *par)
{
	if (current_par_valid) {
		*par = current_par;
	} else {
		fbhw->decode_var (&cvision_fb_predefined [cvision_fb_mode], par);
	}
}

void cvision_fb_set_par (struct cvision_fb_par *par)
{
	current_par = *par;
	current_par_valid = 1;
}

static int do_fb_set_var (struct fb_var_screeninfo *var, int isactive)
{
	int err, activate;
	struct cvision_fb_par par;
	
	if ((err = fbhw->decode_var (var, &par))) {
		return (err);
	}
	
	activate = var->activate;
	
	if ((var->activate & FB_ACTIVATE_MASK) == FB_ACTIVATE_NOW && isactive) {
		cvision_fb_set_par (&par);
	}
	
	fbhw->encode_var (var, &par);
	var->activate = activate;
	
	return (0);
}

void cvision_fb_set_disp (int con)
{
	struct fb_fix_screeninfo fix;
	
	cvision_fb_get_fix (&fix, con);
	if (con == -1) {
		con = 0;
	}
	
	disp [con].screen_base = (u_char *) fix.smem_start;
	disp [con].visual = fix.visual;
	disp [con].type = fix.type;
	disp [con].type_aux = fix.type_aux;
	disp [con].ypanstep = fix.ypanstep;
	disp [con].ywrapstep = fix.ywrapstep;
	disp [con].line_length = fix.line_length;
	disp [con].can_soft_blank = 1;
	disp [con].inverse = cvision_fb_inverse;
}

static int get_video_mode (const char *name)
{
	int i;
	
	for (i = 1; i < NUM_PREDEF_MODES; i++) {
		if (!strcmp (name, cvision_fb_modenames [i])) {
			return (i);
		}
	}
	
	return (0);
}

static struct fb_cmap *get_default_colormap (int bits_per_pixel)
{
	/* should have a switch here for 8bpp, 16bpp, 24/32bpp */
	return (&default_16_colors);
}

static int do_fb_get_cmap (struct fb_cmap *cmap, struct fb_var_screeninfo *var, int kspc)
{
	int i, start;
	u_short *red, *green, *blue, *transp;
	u_int hred, hgreen, hblue, htransp;
	
	   red = cmap->red;
	 green = cmap->green;
	  blue = cmap->blue;
	transp = cmap->transp;
	 start = cmap->start;
	
	if (start < 0) {
		return (-EINVAL);
	}
	
	for (i = 0; i < cmap->len; i++) {
		if (fbhw->getcolreg (start++, &hred, &hgreen, &hblue, &htransp)) {
			return (0);
		}
		
		   hred = CNVT_FROMHW (hred, var->red.length);
		 hgreen = CNVT_FROMHW (hgreen, var->green.length);
		  hblue = CNVT_FROMHW (hblue, var->blue.length);
		htransp = CNVT_FROMHW (htransp, var->transp.length);
		
		if (kspc) {
			 *red = hred;
			*green = hgreen;
			 *blue = hblue;
			
			if (transp) {
				*transp = htransp;
			}
		} else {
			put_user (hred, red);
			put_user (hgreen, green);
			put_user (hblue, blue);
			
			if (transp) {
				put_user (htransp, transp);
			}
		}
		
		red++; green++;  blue++;
		
		if (transp) {
			transp++;
		}
	}
	
	return (0);
}

static int do_fb_set_cmap (struct fb_cmap *cmap, struct fb_var_screeninfo *var, int kspc)
{
	int i, start;
	u_short *red, *green, *blue, *transp;
	u_int hred, hgreen, hblue, htransp;
	
	   red = cmap->red;
	 green = cmap->green;
	  blue = cmap->blue;
	transp = cmap->transp;
	 start = cmap->start;
	 
	 if (start < 0) {
	 	return (-EINVAL);
	 }
	 
	 for (i = 0; i < cmap->len; i++) {
	 	if (kspc) {
	 		   hred = *red;
	 		 hgreen = *green;
	 		  hblue = *blue;
	 		htransp = transp ? *transp : 0;
	 	} else {
	 		get_user (hred, red);
	 		get_user (hgreen, green);
	 		get_user (hblue, blue);
	 		
	 		if (transp) {
	 			get_user (htransp, transp);
	 		} else {
	 			htransp = 0;
	 		}
	 	}
	 	
	 	   hred = CNVT_TOHW (hred, var->red.length);
	 	 hgreen = CNVT_TOHW (hgreen, var->green.length);
	 	  hblue = CNVT_TOHW (hblue, var->blue.length);
	 	htransp = CNVT_TOHW (htransp, var->transp.length);
	 	
	 	red++; green++; blue++;
	 	
	 	if (transp) {
	 		transp++;
	 	}
	 	
	 	if (fbhw->setcolreg (start++, hred, hgreen, hblue, htransp)) {
	 		return (0);
	 	}
	 }
	 
	 return (0);
}

static void do_install_cmap (int con)
{
	if (con != currcon) {
		return;
	}
	
	if (disp [con].cmap.len) {
		do_fb_set_cmap (&disp [con].cmap, &disp [con].var, 1);
	} else {
		do_fb_set_cmap (get_default_colormap (disp [con].var.bits_per_pixel), &disp [con].var, 1);
	}
}

static void memcpy_fs (int fsfromto, void *to, void *from, int len)
{
	switch (fsfromto) {
		case 0:
			memcpy (to, from, len);
			return;
		
		case 1:
			copy_from_user (to, from, len);
			return;
		
		case 2:
			copy_to_user (to, from, len);
			return;
		
		default:
			return;
	}
}

static void copy_cmap (struct fb_cmap *from, struct fb_cmap *to, int fsfromto)
{
	int size;
	int tooff = 0, fromoff = 0;
	
	if (to->start > from->start) {
		fromoff = to->start - from->start;
	} else {
		tooff = from->start - to->start;
	}
	
	size = to->len - tooff;
	
	if (size > from->len - fromoff) {
		size = from->len - fromoff;
	}
	
	if (size < 0) {
		return;
	}
	
	size *= sizeof (u_short);
	
	memcpy_fs (fsfromto, to->red + tooff, from->red + fromoff, size);
	memcpy_fs (fsfromto, to->green + tooff, from->green + fromoff, size);
	memcpy_fs (fsfromto, to->blue + tooff, from->blue + fromoff, size);
	
	if (from->transp && to->transp) {
		memcpy_fs (fsfromto, to->transp + tooff, from->transp + fromoff, size);
	}
}

static int alloc_cmap (struct fb_cmap *cmap, int len, int transp)
{
	int size = len * sizeof (u_short);
	
	if (cmap->len != len) {
		if (cmap->red) {
			kfree (cmap->red);
		}
		
		if (cmap->green) {
			kfree (cmap->green);
		}
		
		if (cmap->blue) {
			kfree (cmap->blue);
		}
		
		if (cmap->transp) {
			kfree (cmap->transp);
		}
		
		cmap->red = cmap->green = cmap->blue = cmap->transp = NULL;
		cmap->len = 0;
		
		if (!len) {
			return (0);
		}
		
		if (!(cmap->red = kmalloc (size, GFP_ATOMIC))) {
			return (-1);
		}
		
		if (!(cmap->green = kmalloc (size, GFP_ATOMIC))) {
			return (-1);
		}
		
		if (!(cmap->blue = kmalloc (size, GFP_ATOMIC))) {
			return (-1);
		}
		
		if (transp) {
			if (!(cmap->transp = kmalloc (size, GFP_ATOMIC))) {
				return (-1);
			}
		} else {
			cmap->transp = NULL;
		}
	}
	
	cmap->start = 0;
	cmap->len = len;
	copy_cmap (get_default_colormap (len), cmap, 0);
	
	return (0);
}


/* ----------------------------------------------------------- */
